/**
 * A heavily re-written UTConsolePlayerInput class that is friendly with PC play ~fox
 * Based off and may contain code provided by and Copyright 1998-2008 Epic Games, Inc. All Rights Reserved.
 */
class foxPlayerInput extends UTPlayerInput within UTPlayerController;

var bool bLastGamepadUse;
var float LastGamepadUseUpdateTime;
var float PendingShowMapControlTime;
var float PendingDropWeaponControlTime;

var float PitchSensitivityScale;

var bool bFineAiming;
var config float FineAimSensitivityScale;

var float QuickTurnThreshold;
var float QuickTurnScale;

var bool bAutoCenterPitch;
var float LastLookTime;
var config float AutoCenterDelay;
var config float AutoPitchCenterSpeed;

var config float Dodge_Threshold;

var bool bEmulateMouse;
var config int bSteamPad[4];
var config float DeadzoneMovement[4];
var config float DeadzoneAiming[4];

//fox: SetSteamPad console function
exec function SetSteamPad(int i, bool bNewSteamPad)
{
	local UTPlayerController PC;

	bSteamPad[i] = int(bNewSteamPad);
	foreach LocalPlayerControllers(class'UTPlayerController', PC) {
		if (LocalPlayer(PC.Player).ControllerId == i) {
			foxPlayerInput(PC.PlayerInput).bEmulateMouse = bNewSteamPad;
			foxPlayerInput(PC.PlayerInput).bSteamPad[i] = bSteamPad[i];
			PC.SetTimer(0.01, false, 'LoadSettingsFromProfile');
			break;
		}
	}
	SaveConfig();
}

//fox: Re-written input processing (view accel, etc.)
function PreProcessInput(float DeltaTime)
{
	local bool bReallyUsingGamepad;

	//Act like a mouse if we're emulating a mouse on a SteamPad
	//Note: We keep specifically keep autoaim later since it's not perfect, and real mouse movements are disabled
	//if (aMouseX != 0f || aMouseY != 0f || aTurn != 0f || aLookUp != 0f)
	//	ClientMessage(aMouseX @ aMouseY @ aTurn @ aLookUp);
	if (bEmulateMouse) {
		if (aMouseX != 0f && aMouseY != 0f
		&& WorldInfo.TimeSeconds - LastGamepadUseUpdateTime > 2.0) {
			LastGamepadUseUpdateTime = WorldInfo.TimeSeconds;
			ClientMessage("Warning: bSteamPad enabled! Mouse input ignored... Use 'SetSteamPad" @ LocalPlayer(Player).ControllerId @ "False' to disable!");
		}
		if (Abs(RawJoyLookRight) >= 0.9999f)
			aTurn *= QuickTurnScale;
		if (Abs(RawJoyLookUp) >= 0.9999f)
			aLookUp *= QuickTurnScale;
		aMouseX = aTurn * 20f;
		aMouseY	= aLookUp * -20f;
		aTurn = 0f;
		aLookUp = 0f;
	}

	//Do the usual processing (controller sensitivity is set here)
	Super.PreProcessInput(DeltaTime);

	//Gamepad specific things
	if (bUsingGamepad) {
		//Always zero DoubleClickTime for Gamepad input this tick -- TODO hook up a config for this?
		DoubleClickTime = 0.0;

		//We're taking the full input from pitch for accurate stick measurement, so scale pitch here
		aLookUp *= PitchSensitivityScale;

		//Fine aiming
		if (bFineAiming)
			ApplyFineAiming();

		//Quick turning (instant view acceleration)
		ApplyQuickTurning();

		//View centering
		if (bAutoCenterPitch)
			ApplyViewAutoPitchCentering(DeltaTime);
	}

	//Possibly UpdateGamepadUse (allow autoaim) if we've been using gamepad for 4 seconds - cache bUsingGamepad so we don't throw ConsoleCommands every check
	//Also consider aMouse inputs as non-gamepad use
	bReallyUsingGamepad = bUsingGamepad && (bEmulateMouse || (aMouseX == 0f && aMouseY == 0f));
	if (bReallyUsingGamepad != bLastGamepadUse && (!bReallyUsingGamepad || WorldInfo.TimeSeconds - LastGamepadUseUpdateTime > 4.0)) {
		ConsoleCommand("UpdateGamepadUse" @ bReallyUsingGamepad);
		LastGamepadUseUpdateTime = WorldInfo.TimeSeconds;
		bLastGamepadUse = bReallyUsingGamepad;
	}

	//Check for certain pending hold-button controls
	if (PendingShowMapControlTime > 0 && WorldInfo.TimeSeconds > PendingShowMapControlTime) {
		ConsoleCommand("ShowMap");
		PendingShowMapControlTime = 0;
	}
	else if (PendingDropWeaponControlTime > 0 && WorldInfo.TimeSeconds > PendingDropWeaponControlTime) {
		ConsoleCommand("DropWeapon");
		PendingDropWeaponControlTime = 0;
	}

	//Adjust sensitivity based on zoom factor, not raw FOV values (hence bEnableFOVScaling=False)
	AdjustZoomSensitivity();
}

//fox: Adjust sensitivity based on zoomed FOV (versus raw FOV, which makes our Hor+ FOV erronously modify sensitivity)
function AdjustZoomSensitivity()
{
	local float FOVScale;

	//Adjust only if currently zoomed in
	if (Outer.FOVAngle < Outer.OnFootDefaultFOV) {
		FOVScale = Outer.FOVAngle / Outer.OnFootDefaultFOV;

		//Mouse
		aMouseX *= FOVScale;
		aMouseY *= FOVScale;

		//Gamepad
		aLookUp *= FOVScale;
		aTurn *= FOVScale;
	}
}

//fox: Fine-aim sensitivity adjustment
function ApplyFineAiming()
{
	aTurn *= FineAimSensitivityScale;
	aLookUp *= FineAimSensitivityScale;
}

//fox: Re-written view acceleration - I really don't like acceleration over time, so I'm going to just curve sensitivity instead (and handle being JAMMED to the edge as you are spazzing out)
function ApplyQuickTurning()
{
	aTurn *= Abs(aTurn);
	aLookUp *= Abs(aLookUp);

	if (!bFineAiming && Abs(RawJoyLookRight) + Abs(RawJoyLookUp) > QuickTurnThreshold) {
		aTurn *= QuickTurnScale;
		aLookUp *= QuickTurnScale;
	}
}

//fox: Re-written look centering
function ApplyViewAutoPitchCentering(float DeltaTime)
{
	local float AmountToAdjustPitch, CurrentPitch;

	//If we're aiming, looking vertically, standing still, firing, zoomed, or changing elevation, don't do anything
	if (Abs(RawJoyLookUp) > 0.0 || (Pawn != None && Pawn.IsFiring()))
		LastLookTime = WorldInfo.TimeSeconds;

	if (bFineAiming || Abs(RawJoyUp) + Abs(RawJoyRight) == 0.0
	|| Pawn == None || Pawn.Physics == PHYS_Falling
	|| UTWeapon(Pawn.Weapon) == None || UTWeapon(Pawn.Weapon).GetZoomedState() != ZST_NotZoomed)
		return;

	if (WorldInfo.TimeSeconds - LastLookTime > AutoCenterDelay) {
		CurrentPitch = float(NormalizeRotAxis(Rotation.Pitch));
		if (CurrentPitch == 0.0)
			return;

		AmountToAdjustPitch = FInterpTo(0.0, CurrentPitch, DeltaTime, AutoPitchCenterSpeed);
		if (bInvertMouse)
			AmountToAdjustPitch *= -1;
		aLookUp += AmountToAdjustPitch;
	}
}

//fox: New "fine aim" option for gamepads - adjusts sensitivity for easier shots
exec function BeginFineAim()
{
	local UTHUD MyUTHUD;

	//No gamepad? Already fine-aiming? Bail here
	if (!bUsingGamepad || bFineAiming)
		return;

	bFineAiming = true;

	MyUTHUD = UTHUD(Outer.myHUD);
	if (MyUTHUD != None)
		MyUTHUD.ConfiguredCrosshairScaling *= 0.75;
}
exec function EndFineAim()
{
	local UTHUD MyUTHUD;

	if (!bFineAiming)
		return;

	bFineAiming = false;

	MyUTHUD = UTHUD(Outer.myHUD);
	if (MyUTHUD != None)
		MyUTHUD.ConfiguredCrosshairScaling /= 0.75;
}

//fox: "SmartJump" is now duck / dodging when moving
exec function SmartJump()
{
	//Dodge or possibly duck if our feet are planted firmly on the ground
	if (Pawn == None || (Pawn.Physics != PHYS_Swimming && Pawn.Physics != PHYS_Flying)) {
		//Handle dodging in four side quadrants - up/down first, then right/left (diagonal clicks in the deadzone go to right/left)
		if (RawJoyUp - Abs(RawJoyRight) > Dodge_Threshold)
			ForcedDoubleClick = DCLICK_Forward;
		else if (RawJoyUp + Abs(RawJoyRight) < -Dodge_Threshold)
			ForcedDoubleClick = DCLICK_Back;
		else if (RawJoyRight > Dodge_Threshold)
			ForcedDoubleClick = DCLICK_Right;
		else if (RawJoyRight < -Dodge_Threshold)
			ForcedDoubleClick = DCLICK_Left;
		else if (!bHoldDuck) {
			Duck();
			bHoldDuck = true;
			return;
		}
	}

	//Unduck if anything other than ducking happened
	bHoldDuck = false;
	UnDuck();
}

//fox: Handle ForcedDoubleClicks
function Actor.EDoubleClickDir CheckForDoubleClickMove(float DeltaTime)
{
	local EDoubleClickDir TempDir;

	if (ForcedDoubleClick != DCLICK_None) {
		TempDir = ForcedDoubleClick;
		DoubleClickDir = DCLICK_None;
		ForcedDoubleClick = DCLICK_None;
		return TempDir;
	}

	return Super.CheckForDoubleClickMove(DeltaTime);
}

defaultproperties
{
	bEnableFOVScaling=False
	bLastGamepadUse=False
	bUsingGamepad=False

	bSteamPad[0]=0
	bSteamPad[1]=0
	bSteamPad[2]=0
	bSteamPad[3]=0
	DeadzoneMovement[0]=0.3
	DeadzoneMovement[1]=0.3
	DeadzoneMovement[2]=0.3
	DeadzoneMovement[3]=0.3
	DeadzoneAiming[0]=0.2
	DeadzoneAiming[1]=0.2
	DeadzoneAiming[2]=0.2
	DeadzoneAiming[3]=0.2

	bFineAiming=False
	LastLookTime=0.0
	PitchSensitivityScale=0.8
	FineAimSensitivityScale=0.6
	bAutoCenterPitch=False
	AutoCenterDelay=1.0
	AutoPitchCenterSpeed=0.01
	QuickTurnThreshold=0.98
	QuickTurnScale=1.0
	Dodge_Threshold=0.025
}
