/**
 * Calculates visible path nodes to every active player (and bot) to allow bots to avoid enemies when pathing to objectives ~fox
 * Based off and may contain code provided by and Copyright 1998-2008 Epic Games, Inc. All Rights Reserved.
 */
class foxAvoidanceMapper extends ReplicationInfo;

struct AvoidanceInfo
{
	var Controller OwningController;
	var array<NavigationPoint> Points;
};
var array<AvoidanceInfo> RouteAvoidanceCache;

var array<LiftExit> LiftExits;

var float RouteAvoidanceBuildInterval;
var float RouteAvoidanceSearchRadius;
var bool bRouteAvoidanceFullTrace;

var array<Controller> ControllerQueue;
var float CalcInterval;

function Setup(float BuildInterval, float SearchRadius, bool bSearchFullTrace)
{
	local LiftExit LE;

	//Check if already initialized (can infer from timers running)
	if (IsTimerActive('CalcVisibility') || IsTimerActive('BuildDynamicCosts'))
		return;

	//`Log("foxLog: Starting AvoidanceMapper on Build Interval" @ BuildInterval @ "Search Radius" @ SearchRadius);
	RouteAvoidanceBuildInterval = BuildInterval;
	RouteAvoidanceSearchRadius = SearchRadius;
	bRouteAvoidanceFullTrace = bSearchFullTrace;
	SetTimer(2.0, false, 'BuildDynamicCosts'); //Wait a second to protect against PRI swapping issues (like from FakeSpectate)

	//Cache all LiftExits for quick access later, as the WorldInfo iteration only needs to be done once per match
	LiftExits.Length = 0;
	foreach WorldInfo.AllNavigationPoints(class'LiftExit', LE)
		if (LE.MyLiftCenter != None && LE.MyLiftCenter.MyLift != None)
			LiftExits.AddItem(LE);
}

function BuildRouteAvoidanceCacheIndices()
{
	local Controller C;
	local byte i;

	foreach WorldInfo.AllControllers(class'Controller', C) {
		if (UTPlayerReplicationInfo(C.PlayerReplicationInfo) != None
		&& UTPlayerReplicationInfo(C.PlayerReplicationInfo).OrdersIndex == 255) {
			RouteAvoidanceCache.Add(1);
			RouteAvoidanceCache[RouteAvoidanceCache.Length - 1].OwningController = C;
			//`Log("foxLog: Build: Adding controller" @ C);
		}
	}

	ControllerQueue.Length = 0;
	for (i = 0; i < RouteAvoidanceCache.Length; i++) {
		C = RouteAvoidanceCache[i].OwningController;
		if (C == None || UTPlayerReplicationInfo(C.PlayerReplicationInfo) == None) {
			//`Log("foxLog: Build: Removing controller" @ C @ "at idx" @ i);
			RouteAvoidanceCache[i].Points.Length = 0;
			RouteAvoidanceCache.Remove(i, 1);
			i--; //Check this one again since we just removed an item
			continue;
		}
		UTPlayerReplicationInfo(C.PlayerReplicationInfo).OrdersIndex = i;
		//`Log("foxLog: Build: Set idx" @ i @ "for" @ C);
		ControllerQueue.AddItem(C);
	}
}

function BuildDynamicCosts()
{
	//`Log("foxLog:" @ WorldInfo.TimeSeconds @ "Firing BuildDynamicCosts");
	BuildRouteAvoidanceCacheIndices();
	CalcInterval = (RouteAvoidanceBuildInterval * (WorldInfo.Game.NumPlayers + WorldInfo.Game.NumBots) / 8) / float(ControllerQueue.Length + 1);
	SetTimer(CalcInterval, false, 'CalcVisibility');
}

function CalcVisibility()
{
	local NavigationPoint N;
	local Controller C;
	local Vector HitLocation, HitNormal, CalcPos;
	local Actor ObjActor;
	local byte i;

	if (ControllerQueue.Length < 1) {
		BuildDynamicCosts();
		return;
	}

	SetTimer(CalcInterval, false, 'CalcVisibility');

	i = ControllerQueue.Length - 1;
	C = ControllerQueue[i];
	ControllerQueue.Length = i;

	if (C == None || C.Pawn == None || UTPlayerReplicationInfo(C.PlayerReplicationInfo) == None)
		return;

	CalcPos = C.Pawn.Location + C.Pawn.GetCollisionHeight() * vect(0, 0, 0.5);
	i = UTPlayerReplicationInfo(C.PlayerReplicationInfo).OrdersIndex; //Hijacked for RouteAvoidanceCache index
	//`Log("foxLog:" @ WorldInfo.TimeSeconds @ "Firing CalcVisibility at idx" @ i @ "for" @ C);
	RouteAvoidanceCache[i].Points.Length = 0;

	//Figure out which iterator to use
	if (RouteAvoidanceSearchRadius == -1) {
		//Check here if we should be using Trace or FastTrace to save checking it every loop iteration
		if (bRouteAvoidanceFullTrace) {
			foreach WorldInfo.AllNavigationPoints(class'NavigationPoint', N)
				if (Trace(HitLocation, HitNormal, N.Location, CalcPos, true) == None)
					RouteAvoidanceCache[i].Points.AddItem(N);
		}
		else {
			foreach WorldInfo.AllNavigationPoints(class'NavigationPoint', N)
				if (FastTrace(N.Location, CalcPos))
					RouteAvoidanceCache[i].Points.AddItem(N);
		}
	}
	else {
		if (bRouteAvoidanceFullTrace) {
			foreach WorldInfo.RadiusNavigationPoints(class'NavigationPoint', N, CalcPos, RouteAvoidanceSearchRadius)
				if (Trace(HitLocation, HitNormal, N.Location, CalcPos, true) == None)
					RouteAvoidanceCache[i].Points.AddItem(N);
		}
		else {
			foreach WorldInfo.RadiusNavigationPoints(class'NavigationPoint', N, CalcPos, RouteAvoidanceSearchRadius)
				if (FastTrace(N.Location, CalcPos))
					RouteAvoidanceCache[i].Points.AddItem(N);
		}
	}

	//If we're a bot, update its PathingUpdateInterval to match ours; if we're a player, find a path so bots can inspect our RouteCache
	if (foxBot(C) != None)
		foxBot(C).ForceRefreshRouteInterval = CalcInterval;
	else if (UTPlayerController(C) != None) {
		if (UTPlayerController(C).LastAutoObjective == None) {
			ObjActor = UTGame(WorldInfo.Game).GetAutoObjectiveFor(UTPlayerController(C));
			UTPlayerController(C).LastAutoObjective = (ObjActor != None) ? ObjActor : C.FindRandomDest();
		}
		if (C.FindPathToward(UTPlayerController(C).LastAutoObjective) != None)
			C.MoveTarget = C.RouteCache[0];
	}
}

function Destroyed()
{
	local Mutator M;
	local foxBot B;
	local Controller C;

	if (IsTimerActive('CalcVisibility'))
		ClearTimer('CalcVisibility');
	if (IsTimerActive('BuildDynamicCosts'))
		ClearTimer('BuildDynamicCosts');

	for (M = WorldInfo.Game.BaseMutator; M != None; M = M.NextMutator)
		if (UTMutator_foxAI(M) != None)
			UTMutator_foxAI(M).AvoidanceMapper = None;
	foreach WorldInfo.AllControllers(class'foxBot', B)
		B.AvoidanceMapper = None;
	foreach WorldInfo.AllControllers(class'Controller', C)
		if (UTPlayerReplicationInfo(C.PlayerReplicationInfo) != None)
			UTPlayerReplicationInfo(C.PlayerReplicationInfo).OrdersIndex = 255;

	//`Log("foxLog: AvoidanceMapper is dieing oh noes! It's okay this was probably supposed to happen.");
	Super.Destroyed();
}

defaultproperties
{
	NetUpdateFrequency=1
}
